package Application::ConfigProcessor;

use 5.006;
use strict;
use warnings;

use Getopt::Long;
use YAML qw( LoadFile );
use File::Spec;
use File::HomeDir;
use base qw( Exporter );

# According to the Perl Exporter documentation, do not export names by default
# as it pollutes the namespace. Hence, use EXPORT_OK instead of EXPORT.
our @EXPORT_OK = qw( process_configs );

=head1 NAME

Application::ConfigProcessor - Determine highest preference value for user supplied config values.

=head1 VERSION

Version 0.01

=cut

our $VERSION = '0.01';

=head1 SYNOPSIS

This module provides subroutine(s) to aid in processing of user supplied
application values. Values can currently be derived from three sources:

    1) command-line
    2) user specific YAML file
    3) global YAML file
    4) application defaults

An example of its use is:

    use Application::ConfigProcessor qw( process_configs );

    my $configs = process_configs(
        'application-defaults' => {
            'database-username'  => 'app_user',
            'database-password'  => 'CHANGEME',
            'verbose'            => 'no',
            'user-config-file'   => '.application/config.yaml',
            'global-config-file' => '/etc/application/conifg.yaml',
        },
    );

You can then pass your application command-line options:

    --database-username
    --database-password
    --verbose
    --user-config-file
    --global-config-file

Alternatively, you can also specify default values in YAML config files.
The user-config-file key is not valid to use in the actual user config YAML
file:

    database-username: mikey
    database-password: helikesit
    verbose: yes
    gloabl-config-file: /opt/application/etc/config.yaml

Similarly, the global-config-file key, nor the user-config-file key, are valid
to use in the global config YAML file:

    database-username: dougy
    database-password: bewareofdogs
    verbose: maybe

The order of precedence for values set in the returned hash reference is:

    1) command-line options
    2) user specific YAML file data
    3) global YAML file data
    4) default hash values passed to subroutine

The following code snippet will show the four config sources showing up in the
resulting hash:

    ---{application}---
    #!/usr/bin/perl

    use Application::ConfigProcessor qw( process_configs );

    my $configs = process_configs(
        'application-defaults' => {
            'database-username'  => 'app_user',
            'database-password'  => 'CHANGEME',
            'database-host'      => 'localhost',
            'verbose'            => 'no',
            'global-config-file' => '/etc/application/conifg.yaml',
            'user-config-file'   => '.application/conifg.yaml',
        },
    );

    print $configs->{'database-username'},"\n";
    print $configs->{'database-password'},"\n";
    print $configs->{'database-host'},"\n";
    print $configs->{'verbose'},"\n";
    print $configs->{'global-config-file'},"\n";
    print $configs->{'user-config-file'},"\n";
    ---{end}---

    ---{/etc/application/conifg.yaml}---
    database-username: mikey
    database-password: helikesit
    database-host: 10.0.0.1
    ---{end}---

    ---{${HOME}/.application/conifg.yaml}---
    database-host: 10.0.0.2
    ---{end}---

    $ ./application --database-password=life --verbose=yes
    mikey
    life
    10.0.0.2
    yes
    /etc/application/conifg.yaml
    .application/config.yaml

=head1 EXPORT

process_configs

=head1 SUBROUTINES/METHODS

=head2 process_configs

Pass a hash reference of keys and default values to this subroutine that
can be given on the command-line. Alternatively, these values can be set in a
YAML config file.

=cut

sub process_configs {
    my %args = (
        'global-config-file-key' => 'global-config-file',
        'user-config-file-key'   => 'user-config-file',
        'application-defaults'   => {},
        @_,
    );

    # --NOTE--  design decision:
    # Since we are dealing with the priority of four (4) different hash sources
    # it would be nice to start with the lowest priority and just continue to
    # overwrite the values for keys as we process hashes until the highest
    # priority hash is reached.
    #
    # HOWEVER!
    #
    # We need to start with our highest priority hash (the CLI options) to determine
    # where our next hash is (possibly) located, as we can change the location of our
    # user and global config file locations on the command line, and change the location
    # of our global config file in the user config file. Thus we need to obtain our CLI
    # options first,

    my $command_line_values = {};

    GetOptions(
        $command_line_values,
        (map { $_.'=s' } keys %{$args{'application-defaults'}}),
    );

    # We want to be able to let the user pick the CLI option name for the global and
    # user config files. Thus, they could deviate from 'global-config-file'
    # and 'user-config-file' in their options hash when they call process_configs.
    my $user_config_file_key = $args{'user-config-file-key'};
    my $user_config_file = defined $command_line_values->{$user_config_file_key} ? $command_line_values->{$user_config_file_key}
                         :                                                         $args{'application-defaults'}{$user_config_file_key}
                         ;

    my $user_config_values = _process_user_config_file($user_config_file);

    my $global_config_file_key = $args{'global-config-file-key'};
    my $global_config_file = defined $command_line_values->{$global_config_file_key} ? $command_line_values->{$global_config_file_key}
                           : defined $user_config_values->{$global_config_file_key}  ? $user_config_values->{$global_config_file_key}
                           :                                                           $args{'application-defaults'}{$global_config_file_key}
                           ;

    my $global_config_values = _process_config_file($global_config_file);

    # Now we should have all four possible config sources:
    # 1) command line values
    # 2) user config file values
    # 3) global config file values
    # 4) application defaults

    # Of course we need to order the hashes in reverse priority, that is, lowest
    # priority first:
    my %configs = (
        %{$args{'application-defaults'}},
        %{$global_config_values},
        %{$user_config_values},
        %{$command_line_values},
    );

    return \%configs;
}

# Internal subroutine to take the (hopefully) relative path string of "user_config_file" and
# determine the user's home directory and then load the YAML configs from that location -
# the concatenation of the home directory and the provided "user_config_file" string.
# NOTE: if the user passes in an absolute path in $user_config_file, it will still
# get concatenated properly by File::Spec->catdir
sub _process_user_config_file {
    my ($user_config_file) = @_;

    if (defined $user_config_file) {
        my $home_directory = File::HomeDir->my_home;
        my $user_config_file_path = File::Spec->catdir($home_directory, $user_config_file);
        return _process_config_file($user_config_file_path);
    }
    else {
        return {};
    }
}

# Internal subroutine to take a string and load the YAML file at that location. NOTE:
# the string does not need to represent an absolute path.
sub _process_config_file {
    my ($config_file) = @_;

    if (defined $config_file) {
        my $config_file_values = {};
        if (-f $config_file) {
            $config_file_values = LoadFile($config_file);
            return $config_file_values;
        }
        else {
            return {};
        }
    }
    else {
        return {};
    }
}

=head1 AUTHOR

Matt Zagrabelny, C<< <mzagrabe at d.umn.edu> >>

=head1 BUGS

Please report any bugs or feature requests to C<mzagrabe at d.umn.edu>.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc Application::ConfigProcessor

=head1 LICENSE AND COPYRIGHT

Copyright © 2013 Regents of the University of Minnesota

This program is free software; you can redistribute it and/or modify it
under the terms of either: the GNU General Public License as published
by the Free Software Foundation; or the Artistic License.

See http://dev.perl.org/licenses/ for more information.

=cut

1; # End of Application::ConfigProcessor
